#include <edapi/journal/file.h>
#include <edapi/journal/entry.h>
#include <edapi/commander.h>
#include <edapi/util.h>

#include <gio/gio.h>
#include <stdio.h>

typedef struct {
    gchar *filename;
    gchar *datetime;
    GDateTime *timestamp;
    gint part;
    GList *entries;
    EDCommander *commander;
    gchar *gameversion;

    EDJournalEntry *first;
    EDJournalEntry *last;
} EDJournalFilePrivate;

struct _EDJournalFile {
    GObject parent;
};

struct _EDJournalFileClass {
    GObjectClass parent_class;
};

G_DEFINE_TYPE_EXTENDED(
    EDJournalFile,
    ed_journal_file,
    G_TYPE_OBJECT,
    0,
    G_ADD_PRIVATE(EDJournalFile)
    );

static void ed_journal_file_dispose(GObject *obj)
{
    EDJournalFile *self = ED_JOURNALFILE(obj);
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);

    g_clear_object(&p->first);
    g_clear_object(&p->last);
    g_clear_object(&p->commander);

    G_OBJECT_CLASS(ed_journal_file_parent_class)->dispose(obj);
}

static void ed_journal_file_finalize(GObject *obj)
{
    EDJournalFile *self = ED_JOURNALFILE(obj);
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);

    g_free(p->filename);
    p->filename = NULL;

    g_free(p->datetime);
    p->datetime = NULL;

    if (p->timestamp != NULL) {
        /* my only pet peeve of glib: you never know if their functions
         * fail on a NULL pointer or not
         */
        g_date_time_unref(p->timestamp);
        p->timestamp = NULL;
    }

    g_free(p->gameversion);
    p->gameversion = NULL;

    g_list_free_full(p->entries, g_object_unref);
    p->entries = NULL;

    G_OBJECT_CLASS(ed_journal_file_parent_class)->finalize(obj);
}

static void ed_journal_file_class_init(EDJournalFileClass *klass)
{
    G_OBJECT_CLASS(klass)->dispose = ed_journal_file_dispose;
    G_OBJECT_CLASS(klass)->finalize = ed_journal_file_finalize;
}

static void ed_journal_file_init(EDJournalFile *self)
{
}

EDJournalFile *ed_journal_file_new(void)
{
    return g_object_new(ED_TYPE_JOURNALFILE, NULL);
}

EDErrorCode ed_journal_file_parse_filename(char const *basename,
                                           gchar **date,
                                           gint *part)
{
    GRegex *new_style = NULL;
    GMatchInfo *matches = NULL;
    gchar *timestamp = NULL;
    gchar *index = NULL;
    EDErrorCode ret = ed_error_invalid;

    new_style = g_regex_new(
        "Journal\\.([\\dT\\-]+)\\.(\\d+)\\.log",
        0, 0, NULL
        );
    goto_if_true(new_style == NULL, done);

    if (g_regex_match(new_style, basename, 0, &matches)) {
        timestamp = g_match_info_fetch(matches, 1);
        index = g_match_info_fetch(matches, 2);

        ret = ed_error_success;
    } else {
        ret = ed_error_invalid;
    }

    g_match_info_unref(matches);
    matches = NULL;

    if (ED_SUCCESS(ret)) {
        if (date != NULL) {
            *date = timestamp;
            /* we gave timestamp to the caller */
            timestamp = NULL;
        }
        if (part != NULL) {
            *part = strtol(index, NULL, 10);
        }
    }

done:

    g_free(timestamp);
    g_free(index);
    g_regex_unref(new_style);

    return ret;
}

static EDErrorCode ed_journal_file_parse_timestamp(EDJournalFile *file)
{
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(file);

    return_if_true(p->datetime == NULL, ed_error_invalid);

    if (strchr(p->datetime, 'T') != NULL) {
        /**
         * new style ISO timestamps
         */
        GTimeZone *utc = g_time_zone_new_utc();
        GDateTime *dt = g_date_time_new_from_iso8601(p->datetime, utc);
        g_time_zone_unref(utc);
        return_if_true(dt == NULL, ed_error_invalid);
        p->timestamp = dt;

        return ed_error_success;
    } else {
        /**
         * old school non-ISO timestamps, used around 2021
         */
        int year = 0, month = 0, day = 0, hour = 0, minute = 0, second = 0;
        int ret = 0;

        ret = sscanf(p->datetime,
                     "%2d%2d%2d%2d%2d%2d",
                     &year, &month, &day,
                     &hour, &minute, &second
            );

        if (ret != 6) {
            return ed_error_invalid;
        }

        /* add base year to years
         */
        year += 2000;

        GDateTime *dt = g_date_time_new_utc(
            year, month, day, hour, minute, second);
        return_if_true(dt == NULL, ed_error_invalid);
        p->timestamp = dt;

        return ed_error_success;
    }

    return ed_error_invalid;
}

static EDErrorCode ed_journal_file_parse_(EDJournalFile *self,
                                          char const *filename)
{
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);
    gchar *basename = NULL;
    EDErrorCode ret = ed_error_internal;

    return_if_true(self == NULL, ed_error_args);
    return_if_true(S_EMPTY(filename), ed_error_args);

    basename = g_path_get_basename(filename);
    goto_if_true(S_EMPTY(basename), done);

    ret = ed_journal_file_parse_filename(basename, &p->datetime, &p->part);
    if (ED_ERROR(ret)) {
        goto done;
    }

    ret = ed_journal_file_parse_timestamp(self);
    if (ED_ERROR(ret)) {
        g_free(p->datetime);
        p->datetime = NULL;
        p->part = 0;
        goto done;
    }

    g_free(p->filename);
    p->filename = g_strdup(filename);

    ret = ed_error_success;

done:

    g_free(basename);

    return ret;
}

static EDErrorCode
ed_journal_file_parse_fileheader(EDJournalFile *self,
                                 EDJournalFilePrivate *p,
                                 EDJournalEntry *e)
{

    gchar const *version = ed_journal_entry_get_string(
        e, "gameversion");

    if (!S_EMPTY(version)) {
        g_free(p->gameversion);
        p->gameversion = g_strdup(version);
    }

    return ed_error_success;
}

static EDErrorCode
ed_journal_file_parse_commander(EDJournalFile *self,
                                EDJournalFilePrivate *p,
                                EDJournalEntry *e)
{

    p->commander = ed_commander_new_from_entry(e);
    return ed_error_success;
}

static EDErrorCode ed_journal_file_read_first_(EDJournalFile *self)
{
    EDErrorCode ret = ed_error_internal;
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);

    GFile *file = NULL;
    GFileInputStream *stream = NULL;
    GDataInputStream *reader = NULL;
    EDJournalEntry *entry = NULL;

    char *line = NULL;
    gsize linelen = 0;

    file = g_file_new_for_path(p->filename);
    goto_if_true(file == NULL, done);

    stream = g_file_read(file, NULL, NULL);
    if (stream == NULL) {
        goto done;
    }

    reader = g_data_input_stream_new(G_INPUT_STREAM(stream));
    if (reader == NULL) {
        goto done;
    }

    while (TRUE) {
        line = g_data_input_stream_read_line_utf8(
            reader, &linelen, NULL, NULL
            );
        goto_if_true(line == NULL, done);

        /* ignore empty lines
         */
        g_strchomp(line);
        if (strlen(line) <= 0) {
            g_free(line);
            line = NULL;
            linelen = 0;
            continue;
        }

        entry = ed_journal_entry_new_parse(line, NULL);

        g_free(line);
        line = NULL;
        linelen = 0;

        if (entry == NULL) {
            continue;
        }

        g_clear_object(&p->first);
        p->first = g_object_ref(entry);

        g_clear_object(&entry);

        break;
    }


    ret = ed_error_success;

done:

    g_clear_object(&reader);
    g_clear_object(&stream);
    g_clear_object(&file);

    return ret;
}

static EDErrorCode ed_journal_file_read_last_(EDJournalFile *self)
{
    EDErrorCode ret = ed_error_internal;
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);

    GFile *file = NULL;
    GFileInputStream *stream = NULL;
    GDataInputStream *reader = NULL;
    EDJournalEntry *entry = NULL;

    goffset end = 0, pos = 0;

    char *line = NULL;
    char *last = NULL;
    gsize linelen = 0;

    file = g_file_new_for_path(p->filename);
    goto_if_true(file == NULL, done);

    stream = g_file_read(file, NULL, NULL);
    if (stream == NULL) {
        goto done;
    }

    reader = g_data_input_stream_new(G_INPUT_STREAM(stream));
    if (reader == NULL) {
        goto done;
    }

    if (!g_seekable_seek(G_SEEKABLE(stream), 0, G_SEEK_END, NULL, NULL)) {
        goto done;
    }

    end = g_seekable_tell(G_SEEKABLE(stream));
    pos = end;

    while (TRUE) {
        /* arbitrary */
        pos = end - 100;
        if (pos <= 0) {
            break;
        }

        if (!g_seekable_seek(G_SEEKABLE(stream), pos, G_SEEK_SET,
                             NULL, NULL)) {
            goto done;
        }

        /* try to read last line if we accidentally seeked over two
         */
        do {
            line = g_data_input_stream_read_line_utf8(
                reader, &linelen, NULL, NULL
                );

            if (line == NULL) {
                break;
            }

            /* ignore empty lines
             */
            g_strchomp(line);
            if (strlen(line) <= 0) {
                g_free(line);
                line = NULL;
                linelen = 0;
                continue;
            }

            g_free(last);
            last = line;
        } while (TRUE);

        if (last == NULL) {
            goto done;
        }

        entry = ed_journal_entry_new_parse(last, NULL);

        g_free(last);
        last = NULL;

        if (entry == NULL) {
            continue;
        }

        g_clear_object(&p->last);
        p->last = g_object_ref(entry);

        g_clear_object(&entry);

        break;
    }

    ret = ed_error_success;

done:

    g_clear_object(&reader);
    g_clear_object(&stream);
    g_clear_object(&file);

    return ret;
}

static EDErrorCode
ed_journal_file_load_(EDJournalFile *self,
                      GError **error,
                      gboolean peek)
{
    EDErrorCode ret = ed_error_internal;
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);

    GFile *file = NULL;
    GFileInputStream *stream = NULL;
    GDataInputStream *reader = NULL;
    EDJournalEntry *entry = NULL;
    GError *e = NULL;
    char *line = NULL;
    gsize linelen = 0;

    file = g_file_new_for_path(p->filename);
    goto_if_true(file == NULL, done);

    stream = g_file_read(file, NULL, &e);
    if (stream == NULL) {
        g_propagate_error(error, e);
        goto done;
    }

    reader = g_data_input_stream_new(G_INPUT_STREAM(stream));
    if (reader == NULL) {
        goto done;
    }

    while (TRUE) {
        line = g_data_input_stream_read_line_utf8(
            reader, &linelen, NULL, &e
            );
        if (line == NULL) {
            if (e != NULL) {
                g_propagate_error(error, e);
                goto done;
            } else {
                break;
            }
        }

        /* ignore empty lines
         */
        g_strchomp(line);
        if (strlen(line) <= 0) {
            g_free(line);
            line = NULL;
            linelen = 0;
            continue;
        }

        entry = ed_journal_entry_new_parse(line, NULL);
        goto_if_true(entry == NULL, done);

        g_free(line);
        line = NULL;
        linelen = 0;

        if (ed_journal_entry_is(entry, ED_JOURNAL_ENTRY_FILEHEADER)) {
            ed_journal_file_parse_fileheader(self, p, entry);
        } else if (ed_journal_entry_is(entry, ED_JOURNAL_ENTRY_COMMANDER)) {
            ed_journal_file_parse_commander(self, p, entry);

            /* if we are just peeking, we stop here
             */
            if (peek) {
                g_clear_object(&entry);
                break;
            }
        }

        if (!peek) {
            p->entries = g_list_append(p->entries, entry);
        } else {
            g_clear_object(&entry);
        }
    }

    ret = ed_error_success;

done:

    g_clear_object(&file);
    g_clear_object(&stream);
    g_clear_object(&reader);

    return ret;
}

EDErrorCode ed_journal_file_open(EDJournalFile *file,
                                 char const *filename,
                                 GError **error)
{
    EDErrorCode r = ed_error_success;

    r = ed_journal_file_parse_(file, filename);
    if (ED_ERROR(r)) {
        return r;
    }

    r = ed_journal_file_load_(file, error, TRUE);
    if (ED_ERROR(r)) {
        return r;
    }

    /* files may be empty */
    ed_journal_file_read_first_(file);
    ed_journal_file_read_last_(file);

    return r;
}

EDErrorCode ed_journal_file_load(EDJournalFile *self, GError **error)
{
    return_if_true(self == NULL, ed_error_args);
    return ed_journal_file_load_(self, error, FALSE);
}

GDateTime *ed_journal_file_get_datetime(EDJournalFile *self)
{
    return_if_true(self == NULL, NULL);
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);
    return p->timestamp;
}

EDCommander *ed_journal_file_get_commander(EDJournalFile *self)
{
    return_if_true(self == NULL, NULL);
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);
    return p->commander;
}

gchar const *ed_journal_file_get_gameversion(EDJournalFile *self)
{
    return_if_true(self == NULL, NULL);
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);
    return p->gameversion;
}

EDJournalEntry *ed_journal_file_get_first(EDJournalFile *self)
{
    return_if_true(self == NULL, NULL);
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);
    return p->first;
}

EDJournalEntry *ed_journal_file_get_last(EDJournalFile *self)
{
    return_if_true(self == NULL, NULL);
    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);
    return p->last;
}

gint ed_journal_file_compare_asc(EDJournalFile *lhs, EDJournalFile *rhs)
{
    return_if_true(lhs == NULL || rhs == NULL, 0);

    GDateTime *ldate = ed_journal_file_get_datetime(lhs);
    GDateTime *rdate = ed_journal_file_get_datetime(rhs);
    return_if_true(ldate == NULL || rdate == NULL, 0);

    return g_date_time_compare(ldate, rdate);
}

gint ed_journal_file_compare_desc(EDJournalFile *lhs, EDJournalFile *rhs)
{
    return_if_true(lhs == NULL || rhs == NULL, 0);

    GDateTime *ldate = ed_journal_file_get_datetime(lhs);
    GDateTime *rdate = ed_journal_file_get_datetime(rhs);
    return_if_true(ldate == NULL || rdate == NULL, 0);

    return g_date_time_compare(rdate, ldate);
}

gboolean ed_journal_file_has_entries_in_range(EDJournalFile *self,
                                              GDateTime *start,
                                              GDateTime *end)
{
    return_if_true(self == NULL, FALSE);

    EDJournalFilePrivate *p = ed_journal_file_get_instance_private(self);

    return_if_true(p->first == NULL || p->last == NULL, FALSE);

    GDateTime *first = ed_journal_entry_get_timestamp(p->first);
    GDateTime *last = ed_journal_entry_get_timestamp(p->last);

    if (g_date_time_compare(start, first) <= 0 &&
        g_date_time_compare(last, end) <= 0) {
        return TRUE;
    }

    if (g_date_time_compare(first, start) <= 0 &&
        g_date_time_compare(start, last) <= 0) {
        return TRUE;
    }

    if (g_date_time_compare(first, end) <= 0 &&
        g_date_time_compare(end, last) <= 0) {
        return TRUE;
    }

    return FALSE;
}
