#include <edapi/journal/journal.h>
#include <edapi/journal/file.h>
#include <edapi/util.h>

typedef struct {
    gchar *location;
    GList *files;
} EDJournalPrivate;

struct _EDJournal {
    GObject parent;
};

struct _EDJournalClass {
    GObjectClass parent_class;
};

G_DEFINE_TYPE_EXTENDED(
    EDJournal,
    ed_journal,
    G_TYPE_OBJECT,
    0,
    G_ADD_PRIVATE(EDJournal)
    );

static void ed_journal_finalize(GObject *obj)
{
    EDJournal *self = ED_JOURNAL(obj);
    EDJournalPrivate *p = ed_journal_get_instance_private(self);

    free(p->location);
    p->location = NULL;

    G_OBJECT_CLASS(ed_journal_parent_class)->finalize(obj);
}

static void ed_journal_class_init(EDJournalClass *klass)
{
    G_OBJECT_CLASS(klass)->finalize = ed_journal_finalize;
}

static EDErrorCode ed_journal_determine_location(EDJournal *self)
{
    gchar *location = NULL;
    gchar *tmp = NULL;

    /**
     * on Windows we have %USERPROFILE% which points to the current
     * users home directory. See if they have a Saved Games folder.
     */
    char const *env1 = getenv("%USERPROFILE%");
    char const *env2 = getenv("USERPROFILE");
    if (!S_EMPTY(env1) || !S_EMPTY(env2)) {
        tmp = g_build_path(
            G_DIR_SEPARATOR_S,
            S_EMPTY(env1) ? env2 : env1,
            "Saved Games",
            "Frontier Developments",
            "Elite Dangerous",
            NULL
            );

        if (g_file_test(tmp, G_FILE_TEST_IS_DIR)) {
            location = tmp;
            tmp = NULL;
        }
    }

    if (S_EMPTY(location)) {
        return ed_error_invalid;
    }

    if (!g_file_test(location, G_FILE_TEST_IS_DIR)) {
        g_free(location);
        return ed_error_invalid;
    }

    ed_journal_set_location(self, location);

    return ed_error_success;
}

static void ed_journal_init(EDJournal *self)
{
    ed_journal_determine_location(self);
}

EDJournal *ed_journal_new(void)
{
    return g_object_new(ED_TYPE_JOURNAL, NULL);
}

EDJournal *ed_journal_new_with_location(gchar const *dir)
{
    EDJournal *j = NULL;
    EDErrorCode ret = ed_error_success;

    j = ed_journal_new();
    return_if_true(j == NULL, NULL);

    ret = ed_journal_set_location(j, dir);
    if (ED_ERROR(ret)) {
        g_clear_object(&j);
        return NULL;
    }

    return j;
}

static void ed_journal_load_files(EDJournal *self)
{
    EDJournalPrivate *p = ed_journal_get_instance_private(self);
    GDir *loc = NULL;
    gchar const *name = NULL;
    EDErrorCode ret = ed_error_success;
    EDJournalFile *journalfile = NULL;

    g_list_free_full(p->files, g_object_unref);
    p->files = NULL;

    loc = g_dir_open(p->location, 0, NULL);
    goto_if_true(loc == NULL, done);

    while ((name = g_dir_read_name(loc)) != NULL) {
        if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0) {
            continue;
        }

        gchar *full = g_build_path(
            G_DIR_SEPARATOR_S,
            p->location,
            name,
            NULL
            );

        if (g_file_test(full, G_FILE_TEST_IS_DIR)) {
            goto next;
        }

        ret = ed_journal_file_parse_filename(name, NULL, NULL);
        goto_if_error(ret, next);

        journalfile = ed_journal_file_new();
        goto_if_true(journalfile == NULL, next);

        ret = ed_journal_file_open(journalfile, full, NULL);
        if (ED_SUCCESS(ret)) {
            p->files = g_list_append(p->files, journalfile);
        }

    next:

        g_free(full);
    }

    p->files = g_list_sort(
        p->files, (GCompareFunc)ed_journal_file_compare_asc);

done:

    if (loc != NULL) {
        g_dir_close(loc);
        loc = NULL;
    }
}

gchar const *ed_journal_get_location(EDJournal *self)
{
    return_if_true(self == NULL, NULL);
    EDJournalPrivate *p = ed_journal_get_instance_private(self);
    return p->location;
}

EDErrorCode ed_journal_set_location(EDJournal *self, gchar const *dir)
{
    return_if_true(self == NULL || dir == NULL, ed_error_args);

    EDJournalPrivate *p = ed_journal_get_instance_private(self);

    if (!g_file_test(dir, G_FILE_TEST_IS_DIR)) {
        return ed_error_invalid;
    }

    g_free(p->location);
    p->location = g_strdup(dir);

    ed_journal_load_files(self);

    return ed_error_success;
}

GList *ed_journal_get_files(EDJournal *self)
{
    return_if_true(self == NULL, NULL);
    EDJournalPrivate *p = ed_journal_get_instance_private(self);
    return p->files;
}
